/* 
 * Java wrapper for GList of strings.
 *
 * The Java-Gnome bindings library is free software distributed under
 * the terms of the GNU Library General Public License version 2.
 * 
 * Author: Dan Bornstein, danfuzz@milk.com
 * Copyright 2000 Dan Bornstein, all rights reserved. 
 */

#include <jni.h>
#include <gtk/gtk.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif



/* reference to the class of instances */
static jclass GListCls;

/* reference to the nativepeer field of instances */
static jfieldID nativepeerFid;



/* ----------------------------------------------------------------------------
 * helper functions 
 */

/* Free the data pointer associated with a GList.
 *
 * data: the pointer to free
 * unused: unused, but required for the g_list_foreach protocol
 */
static void freeGListData (gpointer data, gpointer unused)
{
    g_free(data);
}



/* ----------------------------------------------------------------------------
 * native method declarations
 */

/* Native code for gtk.GList.staticInit0(). */
JNIEXPORT void JNICALL Java_gtk_GListString_staticInit0(JNIEnv *env, 
							jclass cls)
{
    /* guarantee that the class won't get collected */
    (*env)->NewGlobalRef(env, cls);

    GListCls = cls;
    nativepeerFid = (*env)->GetFieldID(env, cls, "nativepeer", "J");
}

/* Native code for gtk.GList.new0(). */
JNIEXPORT void JNICALL Java_gtk_GListString_new0(JNIEnv *env, jobject obj)
{
    jlong peerLong = (jlong) NULL;
    (*env)->SetLongField(env, obj, nativepeerFid, peerLong);
}

/* Native code for gtk.GList.finalize0(). */
JNIEXPORT void JNICALL Java_gtk_GListString_finalize0(JNIEnv *env, jobject obj)
{
    jlong peerLong = (*env)->GetLongField(env, obj, nativepeerFid);
    GList *peer = (GList *) peerLong;

    g_list_foreach (peer, &freeGListData, NULL);
    g_list_free (peer);

    /* paranoia */
    peerLong = (jlong) NULL;
    (*env)->SetLongField(env, obj, nativepeerFid, peerLong);
}

/* Native code for gtk.GList.append0(byte[]). */
JNIEXPORT void JNICALL Java_gtk_GListString_append0(JNIEnv *env, jobject obj, 
						    jbyteArray bytes)
{
    jlong peerLong = (*env)->GetLongField(env, obj, nativepeerFid);
    GList *peer = (GList *) peerLong;
    GList *newpeer;
    jsize len = (*env)->GetArrayLength(env, bytes);
    gchar *newstring = g_malloc(len+1);

    (*env)->GetByteArrayRegion(env, bytes, 0, len, newstring);
    newstring[len] = '\0';
   
    newpeer = g_list_append(peer, newstring);
    if (peer != newpeer)
    {
	peerLong = (jlong) peer;
	(*env)->SetLongField(env, obj, nativepeerFid, newpeer);
    }
}



#ifdef __cplusplus
}
#endif
