/*
 *   Java-Gnome
 *
 *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>
 *
 *   The Java-Gnome bindings library is free software distributed under
 *   the terms of the GNU Library General Public License version 2 
 *
 ***********************************************************************/

import java.io.PrintWriter;
import java.util.*;

public class Argument {
    boolean debug = false;
    String name;
    TypeDefinition type;
    String defaultVal;

    /** Emit C code converting the argument from JNI type (JNI and name
	prefixed by "j_") to a C variable, including "special" conversions 
	(jobject or jstring) */
    public void emitJavaToNativeConversion(PrintWriter pw, String argname) {
	if (type.isString() && type.isArray()) {
	    pw.println("  j" + argname + " = (jstring *)malloc(len" + argname + 
		       "*sizeof(jstring));");
	    pw.println("  " + argname + " = (char **)malloc((len" + argname + 
		       "+1)*sizeof(char *));");
	    pw.println("  for (i" + argname + " = 0; i" + argname + " < len" + 
		       argname + "; i" + argname + "++)");
	    pw.print("    {\n      j" + argname + "[i" + argname + "] = ");
	    pw.println("(*env)->GetObjectArrayElement(env, j_" + name + ", i" + 
		       argname + ");");
	    pw.print("      " + argname + "[i" + argname + "] = (char *)");
	    pw.println("(*env)->GetStringUTFChars(env, j" + argname + "[i" + 
		       argname + "], 0);");
	    pw.println("    }");
	    pw.println("  " + argname + "[i" + argname + "] = NULL;");
	}
	else if (type.isString()) {
	    pw.println("  " + argname + " = " + "(j_" + name + ") ? " + 
		       "(char *)(*env)->GetStringUTFChars(env, j_" + name + 
		       ", 0)" + ": NULL;");
	}
	else if (type.isPrimitive() && type.isArray()) {
	    String typeArray = type.getPrimitiveJavaType();
	    pw.print("  " + argname + " = (*env)->Get");
	    pw.print(typeArray.substring(0, 1).toUpperCase());
	    pw.print(typeArray.substring(1, typeArray.length()).toLowerCase());
	    pw.println("ArrayElements(env, j_" + name + ", 0);");
	}
	else if (type.isEnumValue()) {
	    pw.println("  " + argname + " = " + "(" + type.getNativeType() +
		       ")j_" + name + ";");
	}
	else if (type.isPrimitive())
	    pw.println("  " + argname + " = " + "(" + type.getNativeType() + 
		       ")j_" + name + ";");
	else if (type.isTime_t()) {
	    pw.println("  cls" + argname + " = (*env)->GetObjectClass(env, j_" + 
		       name + ");");
	    pw.println("  mid" + argname + " = (*env)->GetMethodID(env, cls" + 
		       argname + ", \"getTime\", \"()J\");");
	    pw.println("  j" + argname + " = (*env)->CallLongMethod(env, j_" + 
		       name + ", mid" + argname + ");");
	    pw.println("  " + argname + " = (" + type.getNativeType() + ")" + 
		       "(j" + argname + "/1000);");
	}
	else if (type.isGpointer()) {
	    pw.println("  " + argname + " = jobject_to_gpointer(env, j_" + 
		       name + ");");
	}
	else {
	    pw.println("  cls" + argname + " = (*env)->GetObjectClass(env, j_" + 
		       name + ");");
	    pw.println("  fid" + argname + " = (*env)->GetFieldID(env, cls" + 
		       argname + ", \"nativepeer\", \"J\");");
	    pw.println("  j" + argname + " = (*env)->GetLongField(env, j_" + 
		       name + ", fid" + argname + ");");
	    pw.println("  " + argname + " = *(" + type.getNativeType() + "*)&j" + 
		       argname + ";");
	    pw.println();
	}
    }

    /** Emit C code appropriate for subsequent JNI to C type conversions. */
    public void emitNativeDeclaration(PrintWriter pw, String argname) {
	if (!type.isPrimitive() && !type.isString() && !type.isTime_t() && 
	    !type.isGpointer() && !type.isEnumValue()) {
	    pw.println("  jfieldID fid" + argname + ";");
	    pw.println("  jclass cls" + argname + ";");
	    pw.println("  jlong j" + argname + ";");
	}
	if (type.isArray() && type.getPrimitiveNativeType().equals("char *")) {
	    pw.println("  int i" + argname + ";");
	    pw.print("  jsize len" + argname + " = (*env)->GetArrayLength(env, ");
	    pw.println("j_" + name + ");");
	    pw.println("  jstring* j" + argname + ";");
	    pw.print("  " + type.getNativeType());
	    pw.println(" " + argname + ";");
	}
	else if (type.isTime_t()) {
	    pw.println("  jmethodID mid" + argname + ";");
	    pw.println("  jclass cls" + argname + ";");
	    pw.println("  jlong j" + argname + ";");
	    pw.print("  " + type.getNativeType());
	    pw.println(" " + argname + ";");
	}
	else if (type.isGpointer()) {
	    pw.print("  " + type.getNativeType() + " " + argname + ";");
	}
	else if (type.isEnumValue()) {
	    pw.println("  " + type.getNativeType() + " " + argname + ";");
	}
	else {
	    pw.print("  " + type.getNativeType());
	    pw.println(" " + argname + ";");
	}
    }

    /** Emit clean-up code that should be included before getting back in the Java 
	world. Currently only manages string desallocation and array cleanups */
    public void emitNativeToJavaConversion(PrintWriter pw, String argname) {
	if (type.isString() && type.isArray()) {
	    pw.println("  if(" + argname + ")");
	    // being paranoid is key to survival
	    pw.println("    {");
	    pw.println("      for (i" + argname + " = 0; i" + argname + " < len" + 
		       argname + "; i" + argname + "++)");
	    pw.println("        {");
	    pw.print("          j" + argname);
	    pw.println("[i" + argname + "] = (*env)->NewStringUTF(env, " + 
		       argname + "[i" + argname + "]);");
	    pw.print("          (*env)->ReleaseStringUTFChars(env, j" + argname);
	    pw.println("[i" + argname + "], " + argname + "[i" + argname + "]);");
	    pw.print("          (*env)->SetObjectArrayElement(env, j_" + name);
	    pw.println(", i" + argname + ", (jobject)j" + argname + "[i" + 
		       argname + "]);");
	    pw.println("        }");
	    pw.println("      free(j" + argname + ");");
	    pw.println("      free(" + argname + ");");
	    pw.println("    }");
	}
	else if (type.isString()) {
	    pw.println("  if(" + argname + ")");
	    // being paranoid is key to survival
	    pw.println("      (*env)->ReleaseStringUTFChars(env, j_" + name + 
		       ", " + argname + ");");
	}
	else if (type.isArray() && type.isPrimitive()) {
	    pw.println("  if(" + argname + ")");
	    // being paranoid is key to survival
	    String typeArray = type.getPrimitiveJavaType();
	    pw.print("      (*env)->Release");
	    pw.print(typeArray.substring(0, 1).toUpperCase());
	    pw.print(typeArray.substring(1, typeArray.length()).toLowerCase());
	    pw.println("ArrayElements(env, j_" + name + ", " + argname + ", 0);");
	}
	else if (type.isGpointer()) {
	    pw.println("  if(" + argname + ")");
	    // being paranoid is key to survival
	    pw.println("    free(" + argname + ");");
	}
    }

    public String getDefaultValue() {
	if (defaultVal != null && (defaultVal.equals("TRUE") || 
				   defaultVal.equals("FALSE") || defaultVal.equals("NULL")))
	    return defaultVal.toLowerCase();
	else if (type.isEnumValue()) {
	    if (defaultVal.indexOf("|") != -1) {
		StringTokenizer st = new StringTokenizer(defaultVal, "|");
		String Next = st.nextToken();
		// Do magic here
		String Return = new String("new " + getType().translateNaming() + 
					   "(" + getType().translateNaming() + "._" + Next);
		while (st.hasMoreTokens()) {
		    Next = st.nextToken();
		    // Do some more magic here
		    Return = Return.concat("|" + getType().translateNaming() + 
					   "._" + Next);
		}
		Return = Return.concat(")");
		return Return;
	    }
	    else
		return (defaultVal.substring(0, 1).toUpperCase() + 
			defaultVal.substring(1, 3).toLowerCase() + "." + defaultVal);
	}
	else
	    return defaultVal;
    }
    
    public String getName() {
	return name;
    }
    
    public TypeDefinition getType() {
	return type;
    }
    
    public boolean hasDefaultValue() {
	return (defaultVal != null);
    }
    
    public boolean isReserved(String s) {
	return (s.equals("default"));
    }
    
    public Argument(TypeDefinition t, String n) {
	type = t;
	if (isReserved(n))
	    name = n + '_';
	else
	    name = n;
    }
    
    public void setDefaultValue(String s) {
	//System.err.println("------>Argument default value:"+s);
	defaultVal = s;
    }
    
    /** Returns a string representation for the argument suitable for
	inclusion within a Java method prototype or local variable
	declaration, for instance "int count" or "String user_name" */
    public String toString() {
	return type.getJavaType() + " " + name;
    }
}
