/*
 *   Java-Gnome
 *
 *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>
 *     
 *   The Java-Gnome bindings library is free software distributed under
 *   the terms of the GNU Library General Public License version 2 
 *
 ***********************************************************************/


import java.util.*;
import java.io.*;

/** This class holds an abstract description of methods, and is able to 
    emit both Java wrapper, C headers and C glue code.
    
    Bugs: JNI Name mangling is not implemented. Please avoid
    underscore in method names.  Will eventually be written because
    of the naming changes requested by miguel.
    
    @author Olivier Gutknecht */

public class FieldDefinition {
    String fieldName;
    String nativePreGlue = null;
    String nativeGlue = null;
    String nativePostGlue = null;
    ClassDefinition objectType;
    TypeDefinition fieldType;
    
    public void emitJavaField(PrintWriter pw) {
	if (!fieldType.isPrimitive()) {
	    pw.println("  native long nativeGet" + fieldName + "();");
	    pw.println("");
	    pw.println("  public " + fieldType.getJavaType() + " get" + 
		       fieldName + "() {");
	    pw.print("    return new " + fieldType.getJavaType() + "(");
	    pw.println("nativeGet" + fieldName + "());");
	    pw.println("  }");
	}
    }

    public void emitNativeBody(PrintWriter pw) {
	int i;
	if (nativePreGlue != null) {
	    pw.print(nativePreGlue + '\n');
	}
	
	/* To get the object itself native pointer */
	pw.print("  " + fieldType.getNativeType());
	pw.println(" _result;");
	pw.print("  jlong");
	pw.println(" _jresult;");
	pw.println("  " + objectType.getClassType().getNativeType() + " cptr;");
	pw.println("  jfieldID fid;");
	pw.println("  jclass cls;");
	pw.println("  jlong jptr;");
	
	//      pw.print("  "+fieldType.getJNIType());
	
	pw.println();
	
	/* Getting native object pointer */
	pw.println("  cls = (*env)->GetObjectClass(env, obj);");
	pw.println("  fid = (*env)->GetFieldID(env, cls, \"nativepeer\", \"J\");");
	pw.println("  jptr = (*env)->GetLongField(env, obj, fid);");
	pw.println("  cptr = *(" + objectType.getClassType().getNativeType() + 
		   "*)&jptr;");
	pw.println();
	pw.print("  _result = (" + fieldType.getNativeType() + ")");
	pw.println("((" + objectType.getClassType().getNativeType() + ") cptr)->" + 
		   fieldName + ";");
	pw.println("  *(" + fieldType.getNativeType() + " *)&_jresult = _result;");
	//pw.println("  _jresult = ("+fieldType.getJNIType()+")_result;");
	pw.println("  return _jresult;");
    }

    public void emitNativeComment(PrintWriter pw) {
	pw.println("/*");
	pw.println(" * Field acces to " + fieldName);
	pw.println(" * Type:    " + fieldType);
	pw.println(" * Signature: " + getSignature());
	pw.println(" */");
    }

    public void emitNativeHeader(PrintWriter pw) {
	emitNativeComment(pw);
	if (!fieldType.isPrimitive()) {
	    pw.print("JNIEXPORT jlong"); //+fieldType.getJNIType());
	    pw.print(" JNICALL " + "Java_" + objectType.packageName.toLowerCase() + 
		     "_" + objectType.className + "_nativeGet" + fieldName);
	    pw.print(" (JNIEnv *, ");
	    pw.print("jobject");
	    pw.println(");");
	    pw.println();
	}
    }

    /**
     * Get the value of nativeGlue.
     * @return Value of nativeGlue.
     */
    public String getNativeGlue() {
	return nativeGlue;
    }

    public String getSignature() {
	StringBuffer sb = new StringBuffer("(");
	sb.append(")" + fieldType.getSignature());
	return sb.toString();
    }

    public FieldDefinition(String n, TypeDefinition r, ClassDefinition o) {
	fieldName = n;
	fieldType = r;
	objectType = o;
    }

    public void printNativeFunction(PrintWriter pw) {
	emitNativeComment(pw);
	if (!fieldType.isPrimitive()) {
	    pw.print("JNIEXPORT ");
	    pw.print("jlong JNICALL ");
	    
	    //  pw.print(fieldType.getJNIType()+" JNICALL ");
	    pw.println("Java_" + objectType.packageName.toLowerCase() + "_" + 
		       objectType.className + "_nativeGet" + fieldName);
	    pw.print("  (JNIEnv * env, ");
	    pw.print("jobject obj");
	    pw.println(")");
	    pw.println("{");
	    emitNativeBody(pw);
	    pw.println("}");
	    pw.println();
	}
    }

    public void setNativeGlue(String v) {
	this.nativeGlue = v;
    }

    public void setNativePostGlue(String v) {
	this.nativePostGlue = v;
    }

    /**
     * Set the value of nativeGlue.
     * @param v  Value to assign to nativeGlue.
     */
    public void setNativePreGlue(String v) {
	this.nativePreGlue = v;
    }
}
